@@ -1,15 +1,17 @@
 package Domain::PublicSuffix;
-$Domain::PublicSuffix::VERSION = '0.10';
+$Domain::PublicSuffix::VERSION = '0.16';
 use strict;
 use warnings;
 use base 'Class::Accessor::Fast';
 
-use Data::Validate::Domain ();
 use Domain::PublicSuffix::Default ();
 use File::Spec ();
+use Net::IDN::Encode ();
 
 __PACKAGE__->mk_accessors(qw/
+	use_default
 	data_file
+	allow_unlisted_tld
 	domain_allow_underscore
 	tld_tree
 	error
@@ -27,14 +29,17 @@ Domain::PublicSuffix - Parse a domain down to root
  use Domain::PublicSuffix;
 
  my $suffix = Domain::PublicSuffix->new({
-	 'data_file' => '/tmp/effective_tld_names.dat'
+   'data_file' => '/tmp/effective_tld_names.dat'
  });
  my $root = $suffix->get_root_domain('www.google.com');
  # $root now contains "google.com"
- 
+
  $root = $suffix->get_root_domain('www.google.co.uk');
  # $root now contains google.co.uk
 
+ my $tld = $suffix->tld();
+ # $tld now contains co.uk
+
 =head1 DESCRIPTION
 
 This module utilizes the "effective_tld_names.dat" provided by Mozilla as a way
@@ -95,11 +100,21 @@ Can take a hashref of arguments:
 
 A fully qualified path, to override the effective_tld_names.dat file.
 
+=item use_default 
+
+Use the provided publicsuffix file, do not search for any other files.
+
 =item domain_allow_underscore
 
 A flag to indicate that underscores should be allowed in hostnames
 (contra to the RFCs). Default: undef.
 
+=item allow_unlisted_tld
+
+A flag to indicate that unlisted TLDs should be passed through. This follows
+the spec as listed on publicsuffix.org, but is not how this module works by
+default, or before 0.16. Default: undef
+
 =back
 
 =back
@@ -108,7 +123,7 @@ A flag to indicate that underscores should be allowed in hostnames
 
 sub new {
 	my ( $class, @args ) = @_;
-	
+
 	my $self = $class->SUPER::new(@args);
 
 	# Compatibility fix
@@ -116,8 +131,8 @@ sub new {
 		$self->data_file( $args[0]->{'dataFile'} );
 	}
 
-	$self->_parse_data_file();
-	
+	$self->tld_tree($self->_parse_data_to_tree());
+
 	return $self;
 }
 
@@ -134,105 +149,121 @@ the error accessor with a human-readable error string.
 =cut
 
 sub get_root_domain {
-	my ( $self, $domain ) = @_;
-	
+	my ( $self, $inbound ) = @_;
+
+	unless ($inbound) {
+		$self->error('No input');
+		return;
+	}
+
+	my $domain = lc($inbound);
+
 	# Clear meta properties
 	foreach ( qw/tld suffix root_domain error/ ) {
 		undef( $self->{$_} );
 	}
-	
+
 	# Check if domain is valid
-	unless ( $self->_validate_domain($domain) ) {
+	unless ( _validate_domain($domain) ) {
 		$self->error('Malformed domain');
 		return;
 	}
-	
+
 	my @domain_array = split(/\./, $domain);
 	my $tld = pop(@domain_array);
 	unless ( defined $self->tld_tree->{$tld} ) {
-		$self->error('Invalid TLD');
-		return;
-	}
-	
-	$self->tld($tld);
-	$self->suffix($tld) if ( scalar( keys %{$self->tld_tree->{$tld}} ) == 0 );
-	
-	# Reverse iterate through domain to find effective root
-	my $last = $self->tld_tree->{$tld};
-	my $effective_root = $tld;
-	
-	while ( !$self->suffix and scalar(@domain_array) > 0 ) {
-		my $sub = pop(@domain_array);
-		next if (! defined $sub);
-		
-		# check if $sub.$last is a root
-		if ( defined $last->{$sub} and scalar(keys %{$last->{$sub}}) == 0 ) {
-			$self->suffix( $sub . "." . $effective_root );
-			
-		} elsif ( defined $last->{'*'} ) {
-			# wildcard means everything is an root, but check for exceptions
-			my $exception_flag = 0;
-			foreach my $sub_check (keys %{$last}) {
-				if ($sub_check =~ /^!/) {
-					$sub_check =~ s/!//;
-					if ($sub eq $sub_check) {
-						$exception_flag++;
-					}
-				}
-			}
-			if ( $exception_flag > 0 ) {
-				# This is is not an root, push it back to the domain
-				push(@domain_array, $sub);
-				$self->suffix($effective_root);
+		if ( $self->allow_unlisted_tld ) {
+			$self->tld($tld);
+			$self->suffix($tld);
+			if ( my $next = pop(@domain_array) ) {
+				$self->root_domain( join( '.', $next, $tld ) );
 			} else {
-				$self->suffix(join(".", $sub, $effective_root));
+				$self->root_domain($tld);
 			}
-			
-		} elsif ( defined $last->{'RootEnable'} and !defined $last->{$sub} ) {
-			# we have nothing left in the domain string, check
-			# if the root we have is enough
-			push( @domain_array, $sub );
-			$self->suffix($effective_root);
+			return $self->root_domain;
+		} else {
+			$self->error('Invalid TLD');
+			return;
 		}
-		
-		$effective_root = join( '.', $sub, $effective_root );
-		$last = $last->{$sub};
 	}
-	
-	# Leave if we still haven't found an effective root
-	if ( !$self->suffix ) {
+
+	$self->tld($tld);
+	my $raw_suffix = $self->get_suffix_for_domain($domain);
+
+	# Leave if we still haven't found a suffix
+	if ( !$raw_suffix ) {
 		$self->error('Domain not valid');
 		return;
 	}
-	
-	# Check if we're left with just an root
-	if ( $self->suffix eq $domain ) {
-		$self->error('Domain is already root');
+
+	my $suffix = $raw_suffix;
+	$suffix =~ s/!//g;
+	$self->suffix($suffix);
+
+	# Check if we're left with just a suffix
+	if ( $raw_suffix !~ /!/ and $self->suffix eq $domain ) {
+		$self->error('Domain is already a suffix');
 		return;
 	}
-	
-	# Set root domain to one step below effective root.
-	$self->{'root_domain'} = pop(@domain_array) . "." . $self->suffix;
-	
+
+	# Generate root domain using suffix
+	if ($raw_suffix =~ /!/) {
+		# Exception suffixes are also domains
+		$self->root_domain($suffix);
+	} else {
+		my $root_domain = $domain;
+		$root_domain =~ s/^.*\.(.*?\.$suffix)$/$1/;
+		$self->root_domain($root_domain);
+	}
+
 	return $self->root_domain;
 }
 
-sub _parse_data_file {
-	my ( $self ) = @_;
-	
-	$self->{'tld_tree'} = {};
+sub get_suffix_for_domain {
+    my ( $self, $domain ) = @_;
+
+    my @labels = split( /\./, $domain );
+    my $point = $self->tld_tree;
+    my @suffix;
+    while ( my $label = pop(@labels) ) {
+        # If there is a wildcard here, it is a suffix, except for !exceptions
+        # Theoretically, there would be further processing here for .*.
+        # wildcards, but those have not existed before in the list, so saving
+        # the work until it actually happens.
+        if ( $point->{'*'} ) {
+			my $exception = '!' . $label;
+			if ( $point->{$exception} ) {
+				push( @suffix, $exception );
+				last;
+			}
+        } elsif (!$point->{$label}) {
+            # If we run out of rules at this point, the root is just below here
+            last;
+        }
+    
+        push( @suffix, $label );
+        $point = $point->{$label};
+    }
+    return join('.', reverse(@suffix));
+}
+
+sub _load_data {
+	my ($self) = @_;
+
 	my $data_stream_ref;
-	
+
 	# Find an effective_tld_names.dat file
 	my @tld_lines;
 	my $dat;
-	if ( defined $self->data_file and -e $self->data_file ) {
-		open( $dat, '<', $self->data_file )
+	if ( $self->use_default ) {
+		$data_stream_ref = Domain::PublicSuffix::Default::retrieve();
+	} elsif ( $self->data_file and -e $self->data_file ) {
+		open( $dat, '<:encoding(UTF-8)', $self->data_file )
 			or die "Cannot open \'" . $self->data_file . "\': " . $!;
 		@tld_lines = <$dat>;
 		close($dat);
 		$data_stream_ref = \@tld_lines;
-		
+
 	} else {
 		my @paths = (
 			File::Spec->catdir(File::Spec->rootdir, qw/ usr share publicsuffix /),
@@ -244,7 +275,7 @@ sub _parse_data_file {
 		foreach my $path (@paths) {
 			$path = File::Spec->catfile( $path, "effective_tld_names.dat" );
 			if ( -e $path ) {
-				open( $dat, '<', $path )
+				open( $dat, '<:encoding(UTF-8)', $path )
 					or die "Cannot open \'" . $path . "\': " . $!;
 				@tld_lines = <$dat>;
 				close($dat);
@@ -258,81 +289,96 @@ sub _parse_data_file {
 	unless ( defined $data_stream_ref ) {
 		$data_stream_ref = Domain::PublicSuffix::Default::retrieve();
 	}
-	
-	foreach ( @{$data_stream_ref} ) {
+
+	return $data_stream_ref;
+}
+
+sub _parse_data_to_tree {
+	my ($self) = @_;
+
+    my $data_stream_ref = $self->_load_data();
+    my $tree = {};
+
+    foreach (@{$data_stream_ref}) {
 		chomp;
-		# Remove comments, skip if full line comment, remove if on-line comment
+    
+		# Remove comments, skip if full line comment, remove if inline comment
 		next if ( /^\// or /^[ \t]*?$/ );
-		s/\s.*//;
-		
-		# Break down by dots
-		my @domain_array = split( /\./, $_ );
-		warn scalar @domain_array if ($a);
-		my $last = $self->tld_tree;
-		
-		if (scalar(@domain_array) == 1) {
-			my $sub = pop(@domain_array);
-			next if (!$sub);
-			
-			$last->{$sub} = {} unless ( defined $last->{$sub} );
-			$last->{$sub}->{'RootEnable'} = 1;
-		}
-		
-		# Reverse iterate domain array to build hash tree of tlds
-		while (scalar(@domain_array) > 0) {
-			my $sub = pop(@domain_array);
-			$sub =~ s/\s.*//g;
-			next if (!$sub);
-			
-			$last->{$sub} = {} unless ( defined $last->{$sub} );
-			$last->{$sub}->{'RootEnable'} = 1 if ( scalar @domain_array == 0 );
-			$last = $last->{$sub};
-		}
-	}
+		s/[\s\x{0085}\x{000A}\x{000C}\x{000D}\x{0020}].*//;
+
+		# Parse both unicode and ASCII representations, if needed
+		my @tlds = ($_);
+		my $ascii = Net::IDN::Encode::domain_to_ascii($_);
+		push( @tlds, $ascii ) if ( $_ ne $ascii );
+
+		foreach (@tlds) {
+			# Split domain and convert to a tree
+			my @domain = split( /\./, $_ );
+            my $previous = $tree;
+            while (my $label = pop(@domain)) {
+                $label =~ s/\s.*//;
+                $previous->{$label} ||= {};
+                $previous = $previous->{$label};
+            }
+        }
+    }
+    return $tree;
 }
 
 sub _validate_domain {
-	my ($self, $domain) = @_;
-	
-	my $is_valid = Data::Validate::Domain::is_domain( 
-		$domain, 
-		{
-			'domain_allow_underscore'   => $self->domain_allow_underscore,
-			'domain_private_tld'        => qr/^[a-z0-9]+$/,
-		}
-	);
-	return $is_valid;
+	my ($domain) = @_;
+
+	return ( _validate_length($domain) and _validate_multiple_segments($domain) );
 }
 
+# Domains must have more than one segment with length
+sub _validate_multiple_segments {
+	my ($domain) = @_;
+
+	my @segments = split( /\./, $domain );
+	return unless ( @segments > 1 );
+	foreach my $segment (@segments) {
+		return unless ( length($segment) > 0 );
+	}
+	return 1;
+}
+
+# Domains may not be more than 255 characters in length
+sub _validate_length {
+	my ($domain) = @_;
+
+	my $length = length($domain);
+	return ( $length > 1 and $length <= 255 );
+}
 
 ### Compatibility
 
 sub _parseDataFile {
 	my ($self) = @_;
-	
-	return $self->_parse_data_file();
+
+	$self->tld_tree($self->_parse_data_to_tree());
 }
 sub getRootDomain {
 	my ( $self, $domain ) = @_;
-	
+
 	return $self->get_root_domain($domain);
 }
 
 sub _validateDomain {
 	my ($self, $domain) = @_;
-	
+
 	return $self->_validate_domain($domain);
 }
 
 sub dataFile {
 	my ( $self, $data_file ) = @_;
-	
+
 	return $self->data_file($data_file);
 }
 
 sub rootDomain {
 	my ( $self, $root_domain ) = @_;
-	
+
 	return $self->root_domain($root_domain);
 }
 
@@ -358,7 +404,6 @@ L<http://publicsuffix.org/>
 
 =back
 
-
 =head1 BUGS
 
 Please report any bugs or feature requests to C<bug-domain-publicsuffix at rt.cpan.org>,
@@ -366,12 +411,11 @@ or through the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue
 I will be notified, and then you'll automatically be notified of progress on
 your bug as I make changes.
 
-
 =head1 SUPPORT
 
 You can find documentation for this module with the perldoc command.
 
-	perldoc Domain::PublicSuffix
+  perldoc Domain::PublicSuffix
 
 You can also look for information at:
 
@@ -395,7 +439,6 @@ L<http://search.cpan.org/dist/Domain-PublicSuffix>
 
 =back
 
-
 =head1 CONTRIBUTORS
 
 dkg: Daniel Kahn Gillmor
@@ -404,10 +447,9 @@ gavinc: Gavin Carr
 
 jwieland: Jason Wieland
 
-
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2008-12 Nicholas Melnick, C<nick at abstractwankery.com>.
+Copyright 2008-16 Nicholas Melnick, C<nick at abstractwankery.com>.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
